/*
 * Copyright (C) 2003  Bradley A. Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.BufferedReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 *  Reader for output from the Dalton electronic structure program.
 *
 *  <p>
 *   T. Helgaker, H.J.Aa. Jensen, P. Joergensen, J. Olsen,
 *   K. Ruud,H. Aagren, K.L. Bak, V. Bakken, O. Christiansen,P. Dahle,
 *   E.K. Dalskov, T. Enevoldsen, B. Fernandez,H. Heiberg, H. Hettema,
 *   D. Jonsson, S. Kirpekar, R. Kobayashi, H. Koch, K.V. Mikkelsen,
 *   P. Norman, M.J. Packer, T.A. Ruden, T. Saue, S.P.A. Sauer,
 *   K.O. Sylvester-Hvid, P.R. Taylor, and O. Vahtras: "Dalton release
 *   1.1 (2000), an electronic structure program"
 *
 * @author Bradley A. Smith (bradley@baysmith.com)
 */
class DaltonReader implements MoleculeReader {

  static final double angstromPerBohr = 0.529177249;

  BufferedReader input;
  Hashtable atomTypeMap = new Hashtable();

  public DaltonReader(Reader input) {
    this.input = new BufferedReader(input);
  }

  public void read(Molecule mol) throws Exception {

    // Find energy
    String line;
    while (input.ready()) {
      line = input.readLine();
      if (line.trim().startsWith("Total energy")) {
        mol.energy = toEnergy(line);
      } else if (line.trim().startsWith("Atoms and basis sets")) {
        readAtomTypes();
      } else if (line.trim().startsWith("Molecular geometry")) {
        readCoordinates(mol);
      } else if (line.trim().startsWith("Normal Coordinates")) {
        readFrequencies(mol);
        break;
      }
    }
  }

  double toEnergy(String line) throws Exception {

    double energy = 0.0;

    StringReader sr = new StringReader(line);
    StreamTokenizer token = new StreamTokenizer(sr);
    while (token.nextToken() != StreamTokenizer.TT_EOF) {
      if (token.ttype == StreamTokenizer.TT_NUMBER) {
        energy = token.nval;
        break;
      }
    }
    if (token.ttype == StreamTokenizer.TT_EOF) {
      throw new Exception("Error reading energy");
    }

    return energy;
  }

  void readAtomTypes() throws Exception {

    atomTypeMap.clear();

    String line;
    for (int i = 0; i < 9; ++i) {
      line = input.readLine();
    }
    while (input.ready()) {
      line = input.readLine().trim();
      if (line.startsWith("-----")) {
        break;
      }
      StringTokenizer tokenizer = new StringTokenizer(line);
      if (tokenizer.countTokens() < 3) {
        throw new Exception("Error reading atom types");
      }
      String label = tokenizer.nextToken();
      tokenizer.nextToken();
      Integer atomicNumber = new Integer(tokenizer.nextToken());
      atomTypeMap.put(label, atomicNumber);
    }
  }

  void readCoordinates(Molecule mol) throws Exception {

    mol.removeAtoms();

    String line;
    line = input.readLine();
    line = input.readLine();
    while (input.ready()) {
      line = input.readLine().trim();
      if (line.length() == 0) {
        break;
      }
      Atom a = new Atom();
      StringTokenizer tokenizer = new StringTokenizer(line);
      if (tokenizer.countTokens() < 3) {
        throw new Exception("Error reading coordinates");
      }

      String label = tokenizer.nextToken();
      if (atomTypeMap.containsKey(label)) {
        a.atomicNumber = ((Integer) atomTypeMap.get(label)).intValue();
      } else {

        // Unrecognized atom type
        a.atomicNumber = -1;
      }
      a.position = new Point3d();
      a.position.x = (new Double(tokenizer.nextToken())).doubleValue()
          * angstromPerBohr;
      a.position.y = (new Double(tokenizer.nextToken())).doubleValue()
          * angstromPerBohr;
      a.position.z = (new Double(tokenizer.nextToken())).doubleValue()
          * angstromPerBohr;

      mol.addAtom(a);
    }
  }

  void readFrequencies(Molecule mol) throws Exception {

    mol.removeFrequencies();

    String line;
    line = input.readLine();
    line = input.readLine();
    line = input.readLine();

    while (input.ready()) {
      String headerLine = input.readLine().trim();
      line = input.readLine().trim();
      if (!line.startsWith("-----")) {
        break;
      }

      Vector currentFreqs = new Vector();
      StringTokenizer tokenizer = new StringTokenizer(headerLine);
      int numberOfCurrentFrequencies = tokenizer.countTokens() / 2;
      for (int i = 0; i < numberOfCurrentFrequencies; ++i) {
        tokenizer.nextToken();
        Frequency freq = new Frequency();
        String value = tokenizer.nextToken();
        freq.value = (new Double(value)).doubleValue();
        currentFreqs.addElement(freq);
      }
      Vector3d[] currentVectors = new Vector3d[currentFreqs.size()];

      line = input.readLine();

      for (int i = 0; i < mol.getNumberAtoms(); ++i) {
        line = input.readLine();
        tokenizer = new StringTokenizer(line);
        tokenizer.nextToken();    // ignore first token
        tokenizer.nextToken();    // ignore second token
        for (int j = 0; j < currentFreqs.size(); ++j) {
          currentVectors[j] = new Vector3d();
          currentVectors[j].x =
              (new Double(tokenizer.nextToken())).doubleValue()
              * angstromPerBohr;
        }

        line = input.readLine();
        tokenizer = new StringTokenizer(line);
        tokenizer.nextToken();    // ignore first token
        tokenizer.nextToken();    // ignore second token
        for (int j = 0; j < currentFreqs.size(); ++j) {
          currentVectors[j].y =
              (new Double(tokenizer.nextToken())).doubleValue()
              * angstromPerBohr;
        }

        line = input.readLine();
        tokenizer = new StringTokenizer(line);
        tokenizer.nextToken();    // ignore first token
        tokenizer.nextToken();    // ignore second token
        for (int j = 0; j < currentFreqs.size(); ++j) {
          currentVectors[j].z =
              (new Double(tokenizer.nextToken())).doubleValue()
              * angstromPerBohr;
          ((Frequency) currentFreqs.elementAt(j))
              .addVector(currentVectors[j]);
        }

        // Skip blank line between each atom
        line = input.readLine();
      }
      for (int i = 0; i < currentFreqs.size(); ++i) {
        mol.addFrequency((Frequency) currentFreqs.elementAt(i));
      }

      // Skip blank line between frequency sets
      line = input.readLine();
    }
  }
}
