/*
 * Copyright (C) 2003  Bradley A. Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.BufferedReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Vector;

/**
 * Reads GAMESS output files.
 * @author Bradley A. Smith (bradley@baysmith.com)
 */
class GamessReader implements MoleculeReader {

  static final double angstromPerBohr = 0.529177249;

  BufferedReader input;

  public GamessReader(Reader input) {
    this.input = new BufferedReader(input);
  }

  public void read(Molecule mol) throws Exception {

    // Find energy
    String line;
    while (input.ready()) {
      line = input.readLine();
      if (line.indexOf("TOTAL ENERGY =") >= 0) {
        mol.energy = toEnergy(line);
      } else if (line.indexOf("COORDINATES (BOHR)") >= 0) {
        line = input.readLine();
        readCoordinates(mol);
      } else if (line.indexOf("FREQUENCIES IN CM") >= 0) {
        line = input.readLine();
        line = input.readLine();
        if (line.indexOf("*****") >= 0) {
          line = input.readLine();
          line = input.readLine();
          line = input.readLine();
          line = input.readLine();
          line = input.readLine();
        }
        readFrequencies(mol);
        break;
      }
    }
  }

  double toEnergy(String line) throws Exception {

    double energy = 0.0;

    StringReader sr = new StringReader(line);
    StreamTokenizer token = new StreamTokenizer(sr);
    for (int i = 0; i < 3; ++i) {
      token.nextToken();
    }
    if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
      energy = token.nval;
    } else {
      throw new Exception("Error reading energy");
    }

    return energy;
  }

  void readCoordinates(Molecule mol) throws Exception {

    mol.removeAtoms();

    String line;
    while (input.ready()) {
      line = input.readLine();
      if (line.trim().length() == 0) {
        break;
      }
      Atom a = new Atom();
      StringReader sr = new StringReader(line);
      StreamTokenizer token = new StreamTokenizer(sr);
      token.nextToken();    // ignore first token
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.atomicNumber = (int) token.nval;
        if (a.atomicNumber == 0) {

          // skip dummy atoms
          continue;
        }
      } else {
        throw new Exception("Error reading coordinates");
      }
      a.position = new Point3d();
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.x = token.nval * angstromPerBohr;
      } else {
        throw new Exception("Error reading coordinates");
      }
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.y = token.nval * angstromPerBohr;
      } else {
        throw new Exception("Error reading coordinates");
      }
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.z = token.nval * angstromPerBohr;
      } else {
        throw new Exception("Error reading coordinates");
      }
      mol.addAtom(a);
    }
  }

  void readFrequencies(Molecule mol) throws Exception {

    mol.removeFrequencies();

    String line;
    line = input.readLine();
    while (line.indexOf("FREQUENCY:") >= 0) {
      StringReader freqValRead = new StringReader(line.substring(16));
      StreamTokenizer token = new StreamTokenizer(freqValRead);

      Vector freqs = new Vector();
      Frequency previousFreq = new Frequency();
      while (token.nextToken() != StreamTokenizer.TT_EOF) {
        if (token.ttype == StreamTokenizer.TT_WORD) {

          // Previous frequency was imaginary. Negate its value.
          previousFreq.value = -previousFreq.value;

          // Read next token for current frequency value.
          token.nextToken();
        }

        Frequency f = new Frequency();
        f.value = token.nval;
        freqs.addElement(f);
        previousFreq = f;
      }
      Frequency[] currentFreqs = new Frequency[freqs.size()];
      freqs.copyInto(currentFreqs);
      Vector3d[] currentVectors = new Vector3d[currentFreqs.length];

      line = input.readLine();
      line = input.readLine();
      for (int i = 0; i < mol.getNumberAtoms(); ++i) {
        line = input.readLine();
        StringReader vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);
        token.nextToken();    // ignore first token
        token.nextToken();    // ignore second token
        token.nextToken();    // ignore third token
        for (int j = 0; j < currentFreqs.length; ++j) {
          currentVectors[j] = new Vector3d();
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            currentVectors[j].x = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
        }

        line = input.readLine();
        vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);
        token.nextToken();    // ignore first token
        for (int j = 0; j < currentFreqs.length; ++j) {
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            currentVectors[j].y = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
        }

        line = input.readLine();
        vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);
        token.nextToken();    // ignore first token
        for (int j = 0; j < currentFreqs.length; ++j) {
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            currentVectors[j].z = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
          currentFreqs[j].addVector(currentVectors[j]);
        }
      }
      for (int i = 0; i < currentFreqs.length; ++i) {
        mol.addFrequency(currentFreqs[i]);
      }
      for (int i = 0; i < 15; ++i) {
        line = input.readLine();
        if (line.indexOf("FREQUENCY:") >= 0) {
          break;
        }
      }
    }
  }
}

