/*
 * Copyright (C) 2003  Bradley A. Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.BufferedReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Vector;

/**
 * @author Bradley A. Smith (bradley@baysmith.com)
 */
class Aces2Reader implements MoleculeReader {

  static final double angstromPerBohr = 0.529177249;

  BufferedReader input;

  public Aces2Reader(Reader input) {
    this.input = new BufferedReader(input);
  }

  public void read(Molecule mol) throws Exception {

    String line;
    while (input.ready()) {
      line = input.readLine();
      if (line.indexOf("E(SCF)") >= 0) {
        mol.energy = toEnergy(line);
      } else if (line.indexOf("Z-matrix   Atomic") >= 0) {
        line = input.readLine();
        line = input.readLine();
        readCoordinates(mol);
      } else if (line.indexOf("Normal Coordinates") >= 0) {
        line = input.readLine();
        line = input.readLine();
        line = input.readLine();
        readFrequencies(mol);
        break;
      }
    }
  }

  double toEnergy(String line) throws Exception {

    double energy = 0.0;

    StringReader sr = new StringReader(line);
    StreamTokenizer token = new StreamTokenizer(sr);
    for (int i = 0; i < 5; ++i) {
      token.nextToken();
    }
    if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
      energy = token.nval;
    } else {
      throw new Exception("Error reading energy");
    }

    return energy;
  }

  void readCoordinates(Molecule mol) throws Exception {

    mol.removeAtoms();

    String line;
    while (input.ready()) {
      line = input.readLine();
      if (line.indexOf("-----") >= 0) {
        break;
      }
      Atom a = new Atom();
      StringReader sr = new StringReader(line);
      StreamTokenizer token = new StreamTokenizer(sr);
      token.nextToken();    // ignore first token
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.atomicNumber = (int) token.nval;
        if (a.atomicNumber == 0) {

          // skip dummy atoms
          continue;
        }
      } else {
        throw new Exception("Error reading coordinates at atomic number");
      }
      a.position = new Point3d();
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.x = token.nval * angstromPerBohr;
      } else {
        throw new Exception("Error reading coordinates at x coordinate");
      }
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.y = token.nval * angstromPerBohr;
      } else {
        throw new Exception("Error reading coordinates at y coordinate");
      }
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.z = token.nval * angstromPerBohr;
      } else {
        throw new Exception("Error reading coordinates at z coordinate");
      }
      mol.addAtom(a);
    }
  }

  void readFrequencies(Molecule mol) throws Exception {

    mol.removeFrequencies();

    String line;
    while (input.ready()) {
      line = input.readLine();
      Vector currentFreqs = new Vector();
      StringReader freqValRead = new StringReader(line.trim());
      StreamTokenizer token = new StreamTokenizer(freqValRead);
      Frequency previousFreq = new Frequency();
      while (token.nextToken() != StreamTokenizer.TT_EOF) {
        if (token.ttype == StreamTokenizer.TT_WORD) {

          // Previous frequency was imaginary. Negate its value.
          previousFreq.value = -previousFreq.value;

          // Read next token for current frequency value.
          token.nextToken();
        }
        Frequency freq = new Frequency();
        freq.value = token.nval;
        previousFreq = freq;
        currentFreqs.addElement(freq);
      }

      line = input.readLine();
      line = input.readLine();
      for (int i = 0; i < mol.getNumberAtoms(); ++i) {
        line = input.readLine();
        StringReader vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);
        token.nextToken();    // ignore first token
        for (int j = 0; j < currentFreqs.size(); ++j) {
          Vector3d v = new Vector3d();
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            v.x = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            v.y = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            v.z = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
          ((Frequency) currentFreqs.elementAt(j)).addVector(v);
        }
      }
      for (int i = 0; i < currentFreqs.size(); ++i) {
        mol.addFrequency((Frequency) currentFreqs.elementAt(i));
      }
      line = input.readLine();
      if (line.indexOf("Gradient vector") >= 0) {
        break;
      }
      line = input.readLine();
      if (line.indexOf("Normal modes") >= 0) {
        break;
      }
    }
  }

}

