/*
 * Copyright (C) 2003  Bradley A. Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Vector;

/**
 *  Reads Hyperchem output files.
 * @author Bradley A. Smith (bradley@baysmith.com)
 */
class HyperchemReader implements MoleculeReader {

  BufferedReader input;

  public HyperchemReader(Reader input) {
    this.input = new BufferedReader(input);
  }

  public void read(Molecule mol) throws Exception {

    String line;
    while (input.ready()) {
      line = input.readLine();
      if (line.indexOf("Total Energy") >= 0) {
        mol.energy = toEnergy(line);
      } else if (line.indexOf("Coordinates(Angstrom)") >= 0) {
        line = input.readLine();
        readCoordinates(mol);
      } else if (line.indexOf("Frequencies of the Normal Modes") >= 0) {
        for (int i = 0; i < 3; ++i) {
          line = input.readLine();
        }
        readFrequencies(mol);
        break;
      }
    }
  }

  double toEnergy(String line) throws Exception {

    double energy = 0.0;

    StringReader sr = new StringReader(line);
    StreamTokenizer token = new StreamTokenizer(sr);
    for (int i = 0; i < 3; ++i) {
      token.nextToken();
    }
    if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
      energy = token.nval;
    } else {
      throw new Exception("Error reading energy");
    }

    return energy;
  }

  void readCoordinates(Molecule mol) throws Exception {

    mol.removeAtoms();

    String line;
    while (input.ready()) {
      line = readLine();
      if (line.trim().length() == 0) {
        break;
      }
      Atom a = new Atom();
      StringReader sr = new StringReader(line);
      StreamTokenizer token = new StreamTokenizer(sr);

      // Ignore first token
      token.nextToken();
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.atomicNumber = (int) token.nval;
      } else {
        throw new Exception("Error reading coordinates");
      }
      // Ignore third token
      token.nextToken();
      a.position = new Point3d();
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.x = token.nval;
      } else {
        throw new Exception("Error reading coordinates");
      }
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.y = token.nval;
      } else {
        throw new Exception("Error reading coordinates");
      }
      if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
        a.position.z = token.nval;
      } else {
        throw new Exception("Error reading coordinates");
      }
      mol.addAtom(a);
    }
  }

  void readFrequencies(Molecule mol) throws Exception {

    mol.removeFrequencies();

    String line;
    line = readLine();
    while (line != null && line.indexOf("Mode") >= 0) {
      line = readLine();
      line = readLine();
      StringReader freqValRead = new StringReader(line.trim());
      StreamTokenizer token = new StreamTokenizer(freqValRead);

      // Ignore first token
      token.nextToken();
      Vector freqs = new Vector();
      while (token.nextToken() != StreamTokenizer.TT_EOF) {
        Frequency f = new Frequency();
        f.value = token.nval;
        freqs.addElement(f);
      }
      Frequency[] currentFreqs = new Frequency[freqs.size()];
      freqs.copyInto(currentFreqs);
      Vector3d[] currentVectors = new Vector3d[currentFreqs.length];

      line = readLine();
      for (int i = 0; i < mol.getNumberAtoms(); ++i) {
        line = readLine();
        // Remove beginning of line
        line = line.substring(8);
        StringReader vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);

        for (int j = 0; j < currentFreqs.length; ++j) {
          currentVectors[j] = new Vector3d();
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            currentVectors[j].x = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
        }

        line = readLine();
        // Remove beginning of line
        line = line.substring(8);
        vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);

        for (int j = 0; j < currentFreqs.length; ++j) {
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            currentVectors[j].y = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
        }

        line = readLine();
        // Remove beginning of line
        line = line.substring(8);
        vectorRead = new StringReader(line);
        token = new StreamTokenizer(vectorRead);

        for (int j = 0; j < currentFreqs.length; ++j) {
          if (token.nextToken() == StreamTokenizer.TT_NUMBER) {
            currentVectors[j].z = token.nval;
          } else {
            throw new Exception("Error reading frequencies");
          }
          currentFreqs[j].addVector(currentVectors[j]);
        }
      }
      for (int i = 0; i < currentFreqs.length; ++i) {
        mol.addFrequency(currentFreqs[i]);
      }
      for (int i = 0; i < 5; ++i) {
        line = readLine();
        if (line == null || (line.trim().length() > 0) || (line.indexOf("Mode") >= 0)) {
          break;
        }
      }
    }
    // Remove last 6 frequencies because they are translational and rotational.
    for (int i = 0; i < 6; ++i) {
      mol.removeFrequency(mol.getNumberFrequencies() - 1);
    }
  }

  private String readLine() throws IOException {

    String line = input.readLine();
    while ((line != null) && (line.length() > 0)
        && Character.isDigit(line.charAt(0))) {
      line = input.readLine();
    }
    return line;
  }
}

