/*
 * Copyright (C) 2003  Bradley A. Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Vector;

/**
 * <P>Animating molecular vibrations.
 * The normal modes are read from a file automatically
 * determined to be from Aces2, ADF, Dalton, Gamess, Gaussian,
 * Jaguar, or Mopac.
 * Separate animation files are written for each vibration.
 * The animation is a simple cosine trajectory of the normal mode
 * vectors. The output files are in XYZ format with normal mode vectors
 * attached to each atom.
 *
 * <H3>Usage</H3>
 * <blockquote>
 *  <code>xvibs [-h] [-version] [-useFreqInName] [-bounce] [-frames n]
 *    [-amplitude f] [-list listFile] file frequency_number</code>
 *
 * <P>Note: xvibs.jar must be in your CLASSPATH.
 * <H4>Examples</H4>
 * <blockquote>
 *  <code>xvibs ch3oh_gam.out all</code><br>
 *  Outputs all vibrations into files named '<code>ch3oh_gam.</code>{001-018}<code>.xmol</code>'
 *
 * <P><code>xvibs ch3oh_gam.out 15</code><br>
 * Outputs only vibration 15 into the file named '<code>ch3oh_gam.015.xmol</code>'
 * </blockquote>
 *
 * By default, XYZ files created are in loop animation.
 * The option <code>-bounce</code> will create bouncing animations.
 * </blockquote>
 *
 * @author Bradley A. Smith (bradley@baysmith.com)
 */
public class xvibs {

  static final String version = "16";

  static boolean debugging = false;

  private static double defaultFreqScale = 0.7;
  private static double defaultVectorScale = 2.0;
  private static boolean defaultByBounce = false;
  private static int defaultNumberFrames = 20;
  private static boolean defaultUseFreqInName = false;

  static double freqScale = defaultFreqScale;
  static double vectorScale = defaultVectorScale;
  static boolean byBounce = defaultByBounce;
  static int numberFrames = defaultNumberFrames;
  static String listFileName = null;
  static boolean useFreqInName = defaultUseFreqInName;

  private static final DecimalFormatSymbols decimalSymbols = new DecimalFormatSymbols(Locale.US);

  public static void main(String[] args) {

    // Outer try block to catch fatal errors
    try {

      // Process arguments
      if (debugging) {
        System.err.print("args: ");
        for (int i = 0; i < args.length; ++i) {
          System.err.print(args[i] + ' ');
        }
        System.out.println();
      }

      int nArg = 0;
      while ((nArg < args.length) && (args[nArg].charAt(0) == '-')) {
        if (args[nArg].equals("--")) {
          ++nArg;
          break;
        } else if (args[nArg].equals("-useFreqInName")) {
          useFreqInName = true;
          if (debugging) {
            System.err.println("useFreqInName mode");
          }
          ++nArg;

        } else if (args[nArg].equals("-bounce")) {
          byBounce = true;
          if (debugging) {
            System.err.println("bounce mode");
          }
          ++nArg;

        } else if (args[nArg].equals("-h")) {
          printUsage(System.out);
          return;

        } else if (args[nArg].equals("-version")) {
          System.out.println("xvibs version " + version);
          return;

        } else if (args[nArg].equals("-frames")) {
          ++nArg;
          if (nArg >= args.length) {
            throw new Error("-frames requires an argument");
          }
          try {
            numberFrames = Integer.parseInt(args[nArg]);
            if (numberFrames <= 0) {
              throw new Error(
                  "invalid -frames argument. Must be an positive integer");
            }

          } catch (NumberFormatException e) {
            throw new Error(
                "invalid -frames argument. Must be an positive integer");
          }
          ++nArg;
          if (debugging) {
            System.err.println(numberFrames + " frames");
          }
        } else if (args[nArg].equals("-amplitude")) {
          ++nArg;
          if (nArg >= args.length) {
            throw new Error("-amplitude requires an argument");
          }
          try {
            freqScale = Double.parseDouble(args[nArg]);
          } catch (NumberFormatException e) {
            throw new Error(
                "invalid -amplitude argument. Must be a floating-point number");
          }
          ++nArg;
          if (debugging) {
            System.err.println(freqScale + " amplitude");
          }
        } else if (args[nArg].equals("-list")) {
          ++nArg;
          if (nArg >= args.length) {
            throw new Error("-list requires an argument");
          }
          listFileName = args[nArg];
          ++nArg;
          if (debugging) {
            System.err.println(listFileName + " list file");
          }
        } else {
          break;
        }
      }

      int freq = 0;
      boolean allFrequencies = false;
      String inputFileName = "";
      if (args.length - nArg < 2) {
        System.out.println("Insufficient arguments. Using interactive mode.");
        System.out.println();
        System.out.println("xvibs version " + version
            + ", Copyright (C) 1999, 2000  Bradley A. Smith");
        System.out.println(
            "xvibs comes with ABSOLUTELY NO WARRANTY; for details type `show w'.");
        System.out.println(
            "This is free software, and you are welcome to redistribute it\nunder certain conditions; type `show c' for details.");
        System.out.println();
        try {
          BufferedReader input =
              new BufferedReader(new InputStreamReader(System.in));
          System.out.print("Input file? ");
          String inputString = input.readLine();
          inputFileName = inputString.trim();
          if (inputFileName.length() == 0) {
            throw new Error("invalid input file name");
          }
          if (inputFileName.equals("show w")) {
            showWarranty(System.out);
            return;
          } else if (inputFileName.equals("show c")) {
            showCopying(System.out);
            return;
          }
          System.out.println("Input from " + inputFileName);

          System.out.print("Vibration number [all]? ");
          inputString = input.readLine();
          if ((inputString.length() == 0) || inputString.equals("all")) {
            allFrequencies = true;
          } else {
            allFrequencies = false;
            try {
              freq = Integer.parseInt(inputString) - 1;
            } catch (Exception ex) {
              freq = -1;

              // Error handled with following if statement
            }
          }
          if (freq < 0) {
            throw new Error(
                "Invalid vibration number. Must be a positive integer.");
          }
          System.out.print("Generating ");
          if (allFrequencies) {
            System.out.println("all frequencies");
          } else {
            System.out.println("frequency " + (freq + 1));
          }

          System.out.print("Number of frames [" + numberFrames + "]? ");
          inputString = input.readLine();
          if (inputString.length() > 0) {
            numberFrames = Integer.parseInt(inputString);
          }
          if (numberFrames <= 0) {
            throw new Error(
                "Invalid number of frames. Must be a positive integer.");
          }
          System.out.println("Generating " + numberFrames + " frames");

          System.out.print("Generate for loop or bounce animation [loop]? ");
          inputString = input.readLine();
          if ((inputString.length() > 0)
              && ((inputString.charAt(0) == 'B')
              || (inputString.charAt(0) == 'b'))) {
            byBounce = true;
          }

          System.out.print("Generating ");
          if (byBounce) {
            System.out.print("bounce");
          } else {
            System.out.print("loop");
          }
          System.out.println(" animations");

          System.out.print("Frequency amplitude [" + freqScale + "]? ");
          inputString = input.readLine();
          if (inputString.length() > 0) {
            freqScale = Double.parseDouble(inputString);
          }
          System.out.println("Amplitude " + freqScale);
        } catch (IOException ex) {
          throw new Error("Error reading user input: " + ex);
        }
      } else {
        inputFileName = args[nArg].trim();
        ++nArg;
        if (args[nArg].equals("all")) {
          allFrequencies = true;
        } else {
          freq = Integer.parseInt(args[nArg]) - 1;
        }
      }

      if (debugging) {
        System.err.println("input file: " + inputFileName);
      }
      if (inputFileName.length() == 0) {
        throw new Error("invalid input file name");
      }

      FileReader input = null;
      try {
        input = new FileReader(inputFileName);
      } catch (FileNotFoundException e) {
        throw new Error("unable to open '" + inputFileName + "' for input");
      }

      String baseOutputName = inputFileName.substring(0,
          inputFileName.lastIndexOf('.'));
      if (debugging) {
        System.err.println("output base: " + baseOutputName);
      }

      MoleculeReader reader = ReaderFactory.createReader(input);

      if (debugging) {
        System.err.print("input file type: ");
        if (reader instanceof Gaussian03Reader) {
          System.err.println("Gaussian03");
        } else if (reader instanceof Gaussian98Reader) {
          System.err.println("Gaussian98");
        } else if (reader instanceof Gaussian94Reader) {
          System.err.println("Gaussian94");
        } else if (reader instanceof Gaussian92Reader) {
          System.err.println("Gaussian92");
        } else if (reader instanceof GamessReader) {
          System.err.println("Gamess");
        } else if (reader instanceof Aces2Reader) {
          System.err.println("Aces2");
        } else if (reader instanceof ADFReader) {
          System.err.println("ADF");
        } else if (reader instanceof MopacReader) {
          System.err.println("Mopac");
        } else if (reader instanceof HyperchemReader) {
          System.err.println("Hyperchem");
        } else if (reader instanceof SpartanReader) {
          System.err.println("Spartan");
        } else {
          System.err.println("unknown");
        }
      }
      if (reader == null) {
        throw new Error("unknown input file type");
      }

      Molecule mol = new Molecule();
      try {
        reader.read(mol);
      } catch (Exception e) {
        System.err.println(e);
        if (debugging) {
          e.printStackTrace();
        }
      }
      if (debugging) {
        System.err.println(mol.getNumberAtoms() + " atoms and "
            + mol.getNumberFrequencies() + " frequencies read.");
        System.err.println("Energy: " + mol.energy);
        System.err.println("Atoms:");
        NumberFormat form = new DecimalFormat("0.000000", decimalSymbols);
        NumberFormat form2 = new DecimalFormat("0.00", decimalSymbols);
        String v;
        for (Enumeration f = mol.getFrequencies(); f.hasMoreElements();) {
          Enumeration e = mol.getAtoms();
          Frequency freqRef = (Frequency) f.nextElement();
          Enumeration vectors = freqRef.getVectors();
          System.err.println(mol.getNumberAtoms());
          System.err.println("Energy: " + mol.energy + "  Freq: "
              + freqRef.value);
          for (; e.hasMoreElements() && vectors.hasMoreElements();) {
            Atom a = (Atom) e.nextElement();
            Vector3d vector = (Vector3d) vectors.nextElement();
            System.err.print(a.atomicNumber);
            System.err.print(' ');
            v = form.format(a.position.x);
            for (int i = 0; i < 10 - v.length(); ++i) {
              System.err.print(' ');
            }
            System.err.print(v);
            v = form.format(a.position.y);
            for (int i = 0; i < 10 - v.length(); ++i) {
              System.err.print(' ');
            }
            System.err.print(v);
            v = form.format(a.position.z);
            for (int i = 0; i < 10 - v.length(); ++i) {
              System.err.print(' ');
            }
            System.err.print(v);
            System.err.print("  ");
            v = form2.format(vector.x);
            for (int i = 0; i < 5 - v.length(); ++i) {
              System.err.print(' ');
            }
            System.err.print(v);
            System.err.print(' ');
            v = form2.format(vector.y);
            for (int i = 0; i < 5 - v.length(); ++i) {
              System.err.print(' ');
            }
            System.err.print(v);
            System.err.print(' ');
            v = form2.format(vector.z);
            for (int i = 0; i < 5 - v.length(); ++i) {
              System.err.print(' ');
            }
            System.err.print(v);
            System.err.println();
          }
        }
      }

      // Create animation files
      if (mol.getNumberFrequencies() <= 0) {
        throw new Error("no frequencies found");
      }

      int startFreq;
      int endFreq;
      if (allFrequencies) {
        startFreq = 0;
        endFreq = mol.getNumberFrequencies();
      } else {
        if (freq > mol.getNumberFrequencies()) {
          System.err.println(
              "Error: requested frequency is greater than number of frequencies found, "
              + mol.getNumberFrequencies());
          return;
        }
        startFreq = freq;
        endFreq = freq + 1;
      }

      NumberFormat form3 = new DecimalFormat("000", decimalSymbols);
      NumberFormat form4 = new DecimalFormat("0000.000", decimalSymbols);
      Vector fileNameList = new Vector();
      for (int freqNumber = startFreq; freqNumber < endFreq; ++freqNumber) {
        String fileName;
        if (useFreqInName) {
          fileName = baseOutputName
              + form4.format(mol.getFrequency(freqNumber).value)
              .replace('.', '_') + ".xmol";
        } else {
          fileName = baseOutputName + form3.format(freqNumber + 1) + ".xmol";
        }
        if (listFileName != null) {
          fileNameList.addElement(fileName);
        }
        PrintWriter output;
        try {
          output = new PrintWriter(new FileWriter(fileName));
        } catch (IOException e) {
          System.err.println(e);
          return;
        }

        // Write vibrational structures.
        Vector vib = mol.vibration(freqNumber, numberFrames, freqScale,
            byBounce);
        for (Enumeration frames = vib.elements();
             frames.hasMoreElements();) {
          Molecule frame = (Molecule) frames.nextElement();

          output.println(frame.getNumberAtoms());
          output.println("Energy: " + frame.energy + "  Freq: "
              + (mol.getFrequency(freqNumber)).value);

          for (int i = 0; i < frame.getNumberAtoms(); i++) {
            Atom atom = frame.getAtom(i);
            writeAtom(output, atom, (mol
                .getFrequency(freqNumber)).getVector(i), vectorScale);
          }
        }
        output.close();
      }

      if (listFileName != null) {
        // Output list of files generated.
        PrintWriter output2;
        try {
          output2 = new PrintWriter(new FileWriter(listFileName));
        } catch (IOException e) {
          System.err.println(e);
          return;
        }
        output2.print("var xyzfiles = \"");
        for (int i = 0; i < fileNameList.size() - 1; ++i) {
          output2.print(fileNameList.elementAt(i));
          output2.print(',');
          if ((i + 1) % 5 == 0) {
            output2.println('"');
            output2.print(" + \"");
          }
        }
        output2.print(fileNameList.elementAt(fileNameList.size() - 1));
        output2.print("\";");
        output2.close();
      }
    } catch (Error e) {
      System.err.println("xvibs: " + e.getMessage());
      printUsage(System.err);
    }
  }

  static void writeAtom(PrintWriter out, Atom a, Vector3d v,
      double scaleVector) {

    // perhaps v's x, y, and z should be set to zero for too small amplitudes
    NumberFormat form = new DecimalFormat("0.000000", decimalSymbols);

    //NumberFormat form2 = new DecimalFormat("0.00", decimalSymbols);
    out.print(padRight(AtomicSymbol.atomicNumberToSymbol(a.atomicNumber), ' ', 2));
    out.print(' ');
    out.print(padLeft(form.format(a.position.x), ' ', 11));
    out.print(' ');
    out.print(padLeft(form.format(a.position.y), ' ', 11));
    out.print(' ');
    out.print(padLeft(form.format(a.position.z), ' ', 11));
    out.print(' ');
    out.print(padLeft(form.format(v.x * scaleVector), ' ', 11));
    out.print(' ');
    out.print(padLeft(form.format(v.y * scaleVector), ' ', 11));
    out.print(' ');
    out.print(padLeft(form.format(v.z * scaleVector), ' ', 11));
    out.println();
  }

  static String padLeft(String in, char ch, int width) {

    StringBuffer buffer = new StringBuffer();
    for (int i = in.length(); i < width; ++i) {
      buffer.append(ch);
    }
    buffer.append(in);
    return buffer.toString();
  }

  public static String padRight(String in, char ch, int width) {
    StringBuffer buffer = new StringBuffer(in);
    for (int i = buffer.length(); i < width; ++i) {
      buffer.append(ch);
    }
    return buffer.toString();
  }

  static void printUsage(PrintStream out) {

    out.println(
        "Usage: java xvibs [-h] [-version] [-useFreqInName] [-bounce] [-frames n] [-amplitude f] [-list listFile] file frequency_number");
    out.println("  -useFreqInName  produce files numbered by frequency value rather");
    out.println("                  than frequency number (default: " + defaultUseFreqInName + ")");
    out.println("  -bounce         produce bouncing animations (default: "
        + defaultByBounce + ")");
    out.println("  -frames         output n frames (default: "
        + defaultNumberFrames + ")");
    out.println("  -amplitude      scale frequencies by f (default: "
        + defaultFreqScale + ")");
    out.println("  -list           output a Javascript list of the generated");
    out.println("                  files to the given listFile (default: null)");
  }

  static void showWarranty(PrintStream output) {

    output.println("                    GNU GENERAL PUBLIC LICENSE");
    output.println("                            NO WARRANTY");
    output.println();
    output.println(
        "  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY");
    output.println(
        "FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN");
    output.println(
        "OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES");
    output.println(
        "PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED");
    output.println(
        "OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF");
    output.println(
        "MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS");
    output.println(
        "TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE");
    output.println(
        "PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,");
    output.println("REPAIR OR CORRECTION.");
    output.println();
    output.println(
        "  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING");
    output.println(
        "WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR");
    output.println(
        "REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,");
    output.println(
        "INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING");
    output.println(
        "OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED");
    output.println(
        "TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY");
    output.println(
        "YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER");
    output.println(
        "PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE");
    output.println("POSSIBILITY OF SUCH DAMAGES.");
  }

  static void showCopying(PrintStream output) {

    output.println("                    GNU GENERAL PUBLIC LICENSE");
    output.println(
        "   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION");
    output.println();
    output.println(
        "  0. This License applies to any program or other work which contains");
    output.println(
        "a notice placed by the copyright holder saying it may be distributed");
    output.println(
        "under the terms of this General Public License.  The \"Program\", below,");
    output.println(
        "refers to any such program or work, and a \"work based on the Program\"");
    output.println(
        "means either the Program or any derivative work under copyright law:");
    output.println(
        "that is to say, a work containing the Program or a portion of it,");
    output.println(
        "either verbatim or with modifications and/or translated into another");
    output.println(
        "language.  (Hereinafter, translation is included without limitation in");
    output.println(
        "the term \"modification\".)  Each licensee is addressed as \"you\".");
    output.println();
    output.println(
        "Activities other than copying, distribution and modification are not");
    output.println(
        "covered by this License; they are outside its scope.  The act of");
    output.println(
        "running the Program is not restricted, and the output from the Program");
    output.println(
        "is covered only if its contents constitute a work based on the");
    output.println(
        "Program (independent of having been made by running the Program).");
    output.println("Whether that is true depends on what the Program does.");
    output.println();
    output.println(
        "  1. You may copy and distribute verbatim copies of the Program's");
    output.println(
        "source code as you receive it, in any medium, provided that you");
    output.println(
        "conspicuously and appropriately publish on each copy an appropriate");
    output.println(
        "copyright notice and disclaimer of warranty; keep intact all the");
    output.println(
        "notices that refer to this License and to the absence of any warranty;");
    output.println(
        "and give any other recipients of the Program a copy of this License");
    output.println("along with the Program.");
    output.println();
    output.println(
        "You may charge a fee for the physical act of transferring a copy, and");
    output.println(
        "you may at your option offer warranty protection in exchange for a fee.");
    output.println();
    output.println(
        "  2. You may modify your copy or copies of the Program or any portion");
    output.println(
        "of it, thus forming a work based on the Program, and copy and");
    output.println(
        "distribute such modifications or work under the terms of Section 1");
    output.println(
        "above, provided that you also meet all of these conditions:");
    output.println();
    output.println(
        "    a) You must cause the modified files to carry prominent notices");
    output.println(
        "    stating that you changed the files and the date of any change.");
    output.println();
    output.println(
        "    b) You must cause any work that you distribute or publish, that in");
    output.println(
        "    whole or in part contains or is derived from the Program or any");
    output.println(
        "    part thereof, to be licensed as a whole at no charge to all third");
    output.println("    parties under the terms of this License.");
    output.println();
    output.println(
        "    c) If the modified program normally reads commands interactively");
    output.println(
        "    when run, you must cause it, when started running for such");
    output.println(
        "    interactive use in the most ordinary way, to print or display an");
    output.println(
        "    announcement including an appropriate copyright notice and a");
    output.println(
        "    notice that there is no warranty (or else, saying that you provide");
    output.println(
        "    a warranty) and that users may redistribute the program under");
    output.println(
        "    these conditions, and telling the user how to view a copy of this");
    output.println(
        "    License.  (Exception: if the Program itself is interactive but");
    output.println(
        "    does not normally print such an announcement, your work based on");
    output.println(
        "    the Program is not required to print an announcement.)");
    output.println();
    output.println(
        "These requirements apply to the modified work as a whole.  If");
    output.println(
        "identifiable sections of that work are not derived from the Program,");
    output.println(
        "and can be reasonably considered independent and separate works in");
    output.println(
        "themselves, then this License, and its terms, do not apply to those");
    output.println(
        "sections when you distribute them as separate works.  But when you");
    output.println(
        "distribute the same sections as part of a whole which is a work based");
    output.println(
        "on the Program, the distribution of the whole must be on the terms of");
    output.println(
        "this License, whose permissions for other licensees extend to the");
    output.println(
        "entire whole, and thus to each and every part regardless of who wrote it.");
    output.println();
    output.println(
        "Thus, it is not the intent of this section to claim rights or contest");
    output.println(
        "your rights to work written entirely by you; rather, the intent is to");
    output.println(
        "exercise the right to control the distribution of derivative or");
    output.println("collective works based on the Program.");
    output.println();
    output.println(
        "In addition, mere aggregation of another work not based on the Program");
    output.println(
        "with the Program (or with a work based on the Program) on a volume of");
    output.println(
        "a storage or distribution medium does not bring the other work under");
    output.println("the scope of this License.");
    output.println();
    output.println(
        "  3. You may copy and distribute the Program (or a work based on it,");
    output.println(
        "under Section 2) in object code or executable form under the terms of");
    output.println(
        "Sections 1 and 2 above provided that you also do one of the following:");
    output.println();
    output.println(
        "    a) Accompany it with the complete corresponding machine-readable");
    output.println(
        "    source code, which must be distributed under the terms of Sections");
    output.println(
        "    1 and 2 above on a medium customarily used for software interchange; or,");
    output.println();
    output.println(
        "    b) Accompany it with a written offer, valid for at least three");
    output.println(
        "    years, to give any third party, for a charge no more than your");
    output.println(
        "    cost of physically performing source distribution, a complete");
    output.println(
        "    machine-readable copy of the corresponding source code, to be");
    output.println(
        "    distributed under the terms of Sections 1 and 2 above on a medium");
    output.println("    customarily used for software interchange; or,");
    output.println();
    output.println(
        "    c) Accompany it with the information you received as to the offer");
    output.println(
        "    to distribute corresponding source code.  (This alternative is");
    output.println(
        "    allowed only for noncommercial distribution and only if you");
    output.println(
        "    received the program in object code or executable form with such");
    output.println("    an offer, in accord with Subsection b above.)");
    output.println();
    output.println(
        "The source code for a work means the preferred form of the work for");
    output.println(
        "making modifications to it.  For an executable work, complete source");
    output.println(
        "code means all the source code for all modules it contains, plus any");
    output.println(
        "associated interface definition files, plus the scripts used to");
    output.println(
        "control compilation and installation of the executable.  However, as a");
    output.println(
        "special exception, the source code distributed need not include");
    output.println(
        "anything that is normally distributed (in either source or binary");
    output.println(
        "form) with the major components (compiler, kernel, and so on) of the");
    output.println(
        "operating system on which the executable runs, unless that component");
    output.println("itself accompanies the executable.");
    output.println();
    output.println(
        "If distribution of executable or object code is made by offering");
    output.println(
        "access to copy from a designated place, then offering equivalent");
    output.println(
        "access to copy the source code from the same place counts as");
    output.println(
        "distribution of the source code, even though third parties are not");
    output.println(
        "compelled to copy the source along with the object code.");
    output.println();
    output.println(
        "  4. You may not copy, modify, sublicense, or distribute the Program");
    output.println(
        "except as expressly provided under this License.  Any attempt");
    output.println(
        "otherwise to copy, modify, sublicense or distribute the Program is");
    output.println(
        "void, and will automatically terminate your rights under this License.");
    output.println(
        "However, parties who have received copies, or rights, from you under");
    output.println(
        "this License will not have their licenses terminated so long as such");
    output.println("parties remain in full compliance.");
    output.println();
    output.println(
        "  5. You are not required to accept this License, since you have not");
    output.println(
        "signed it.  However, nothing else grants you permission to modify or");
    output.println(
        "distribute the Program or its derivative works.  These actions are");
    output.println(
        "prohibited by law if you do not accept this License.  Therefore, by");
    output.println(
        "modifying or distributing the Program (or any work based on the");
    output.println(
        "Program), you indicate your acceptance of this License to do so, and");
    output.println(
        "all its terms and conditions for copying, distributing or modifying");
    output.println("the Program or works based on it.");
    output.println();
    output.println(
        "  6. Each time you redistribute the Program (or any work based on the");
    output.println(
        "Program), the recipient automatically receives a license from the");
    output.println(
        "original licensor to copy, distribute or modify the Program subject to");
    output.println(
        "these terms and conditions.  You may not impose any further");
    output.println(
        "restrictions on the recipients' exercise of the rights granted herein.");
    output.println(
        "You are not responsible for enforcing compliance by third parties to");
    output.println("this License.");
    output.println();
    output.println(
        "  7. If, as a consequence of a court judgment or allegation of patent");
    output.println(
        "infringement or for any other reason (not limited to patent issues),");
    output.println(
        "conditions are imposed on you (whether by court order, agreement or");
    output.println(
        "otherwise) that contradict the conditions of this License, they do not");
    output.println(
        "excuse you from the conditions of this License.  If you cannot");
    output.println(
        "distribute so as to satisfy simultaneously your obligations under this");
    output.println(
        "License and any other pertinent obligations, then as a consequence you");
    output.println(
        "may not distribute the Program at all.  For example, if a patent");
    output.println(
        "license would not permit royalty-free redistribution of the Program by");
    output.println(
        "all those who receive copies directly or indirectly through you, then");
    output.println(
        "the only way you could satisfy both it and this License would be to");
    output.println("refrain entirely from distribution of the Program.");
    output.println();
    output.println(
        "If any portion of this section is held invalid or unenforceable under");
    output.println(
        "any particular circumstance, the balance of the section is intended to");
    output.println(
        "apply and the section as a whole is intended to apply in other");
    output.println("circumstances.");
    output.println();
    output.println(
        "It is not the purpose of this section to induce you to infringe any");
    output.println(
        "patents or other property right claims or to contest validity of any");
    output.println(
        "such claims; this section has the sole purpose of protecting the");
    output.println(
        "integrity of the free software distribution system, which is");
    output.println(
        "implemented by public license practices.  Many people have made");
    output.println(
        "generous contributions to the wide range of software distributed");
    output.println(
        "through that system in reliance on consistent application of that");
    output.println(
        "system; it is up to the author/donor to decide if he or she is willing");
    output.println(
        "to distribute software through any other system and a licensee cannot");
    output.println("impose that choice.");
    output.println();
    output.println(
        "This section is intended to make thoroughly clear what is believed to");
    output.println("be a consequence of the rest of this License.");
    output.println("");
    output.println(
        "  8. If the distribution and/or use of the Program is restricted in");
    output.println(
        "certain countries either by patents or by copyrighted interfaces, the");
    output.println(
        "original copyright holder who places the Program under this License");
    output.println(
        "may add an explicit geographical distribution limitation excluding");
    output.println(
        "those countries, so that distribution is permitted only in or among");
    output.println(
        "countries not thus excluded.  In such case, this License incorporates");
    output.println(
        "the limitation as if written in the body of this License.");
    output.println();
    output.println(
        "  9. The Free Software Foundation may publish revised and/or new versions");
    output.println(
        "of the General Public License from time to time.  Such new versions will");
    output.println(
        "be similar in spirit to the present version, but may differ in detail to");
    output.println("address new problems or concerns.");
    output.println();
    output.println(
        "Each version is given a distinguishing version number.  If the Program");
    output.println(
        "specifies a version number of this License which applies to it and \"any");
    output.println(
        "later version\", you have the option of following the terms and conditions");
    output.println(
        "either of that version or of any later version published by the Free");
    output.println(
        "Software Foundation.  If the Program does not specify a version number of");
    output.println(
        "this License, you may choose any version ever published by the Free Software");
    output.println("Foundation.");
    output.println();
    output.println(
        "  10. If you wish to incorporate parts of the Program into other free");
    output.println(
        "programs whose distribution conditions are different, write to the author");
    output.println(
        "to ask for permission.  For software which is copyrighted by the Free");
    output.println(
        "Software Foundation, write to the Free Software Foundation; we sometimes");
    output.println(
        "make exceptions for this.  Our decision will be guided by the two goals");
    output.println(
        "of preserving the free status of all derivatives of our free software and");
    output.println(
        "of promoting the sharing and reuse of software generally.");
  }
}
